#!/usr/bin/env python3.9

import re
from itertools import takewhile
from os import path
from typing import Dict, List, Optional
import sys

import yaml
import fire


PACK_PATH = [
    'packs/{0}/main.sql.fmt',
    'packs/{0}/{0}.sql.fmt',
]


def format(find, replace):
    found = find
    for k in replace:
        found = re.sub(f'{{{k}}}', replace[k], found)
    return find if find == found else format(found, replace)


def expand_macros(pack_dir, pack):
    return re.sub(
        r"USING TEMPLATE '([a-z-_\.]+)'",
        lambda m: f"AS $$\n{open(path.join(pack_dir, m.group(1))).read()}$$",
        pack,
        flags=re.M,
    )


def load_data_file(data_file, **kwargs) -> List[Dict[str, str]]:
    if data_file is None:
        return [{}]

    with open(f"{data_file}", "r") as f:
        data = yaml.safe_load(f)
        if type(data) is list:
            return data
        else:
            return [data]


def validate_params(params: List[Dict[str, str]], arg: Dict[str, str]) -> bool:
    missing_params = set()
    for param in params:
        name = param['name']
        if name not in arg:
            missing_params.add(name)
    return missing_params


def read_template_header(template):
    lines = template.split("\n")
    head = "\n".join(
        re.sub("^-- ", "", l) for l in takewhile(lambda s: s.startswith("-- "), lines)
    )
    header = yaml.safe_load(head)
    return header['varmap'], header['params']


def print_sql(pack_name, data_file: Optional[str] = None, **kwargs):
    pack_file_options = [pack_name] + [p.format(pack_name) for p in PACK_PATH]
    pack_file = next((f for f in pack_file_options if path.isfile(f)), None)
    if not pack_file:
        print("NO PACK FOUND WITH NAME", pack_name, file=sys.stderr)
        sys.exit(-1)
    pack_dir = path.dirname(pack_file)
    template_text = open(pack_file, 'r').read()

    varmap, params = read_template_header(template_text)
    defaults = {p['name']: p['default'] for p in params if 'default' in p}
    for file_datum in load_data_file(data_file):
        args = defaults | varmap | file_datum | kwargs
        missing_params = validate_params(params, args)
        if not missing_params:
            print(expand_macros(pack_dir, format(template_text, args)))
        else:
            for name in missing_params:
                print("MISSING VALUE FOR", name, file=sys.stderr)


if __name__ == '__main__':
    fire.Fire(print_sql)



from distutils.core import setup

setup(
    name='sqlpack',
    packages=['sqlpack'],
    version='0.1',
    license='OSI Approved :: Apache Software License',
    description='The sqlpack module lets you distribute modular SQL code.',
    author='Siddhil Rathod',
    author_email='siddhi.rathod@snowflake.com',
    url='https://github.com/Snowflake-Labs/sqlpack',
    download_url = 'https://github.com/Snowflake-Labs/sqlpacks/archive/refs/tags/0.1.0.tar.gz',
    scripts=['bin/sqlpack'],
    keywords='SNOWFLAKE PACKS SNOWSQL SNOWSQL-PACKS',
    install_requires=[
        'fire',
        'pyyaml',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.9',
    ],
)



import pytest
from . import pack

varmap = {
    "base_name": "AWS_CLOUDTRAIL_conn_EVENTS",
    "landing_table": "data.{base_name}_CONNECTION",
    "pipe": "data.{base_name}_PIPE",
}


@pytest.mark.parametrize(
    "find,found",
    [
        (
            "base_name",
            "base_name",
        ),  # no replacement occurs for variables without the swiggly {} brackets
        (
            "{not_in_varmap}",
            "{not_in_varmap}",
        ),  # no replacement occurs for string in swiggly{} brackets that is not a key in the varmap
    ],
)
def test_format_no_replacement(find, found):
    result = pack.format(find, varmap)
    assert result == found


@pytest.mark.parametrize(
    "find,found",
    [
        (
            "{base_name}",
            "AWS_CLOUDTRAIL_conn_EVENTS",
        )  # single replacement of a valid key in swiggly brackets by the respective value from the varmap
    ],
)
def test_format_single_replacement(find, found):
    result = pack.format(find, varmap)
    assert result == found


@pytest.mark.parametrize(
    "find,found",
    [
        (
            "{landing_table}",
            "data.AWS_CLOUDTRAIL_conn_EVENTS_CONNECTION",
        )  # recursive replacement of a valid key in swiggly brackets which also contains a valid key that needs to be replaced by the respective value from the varmap
    ],
)
def test_format_recursive_replacement(find, found):
    result = pack.format(find, varmap)
    assert result == found


@pytest.mark.parametrize(
    "find,found",
    [
        (
            "{base_name} is used in {landing_table}",
            "AWS_CLOUDTRAIL_conn_EVENTS is used in data.AWS_CLOUDTRAIL_conn_EVENTS_CONNECTION",
        )  # A single and a recursice replacement in a single line
    ],
)
def test_format_multiple_and_recursive_replacement(find, found):
    result = pack.format(find, varmap)
    assert result == found
